import { getItems, checkContext } from 'feathers-hooks-common';
import copy from 'fast-copy';
import _get from 'lodash/get.js';
import _set from 'lodash/set.js';
import _isEqual from 'lodash/isEqual.js';
import { shouldSkip } from 'feathers-utils';
import transformMustache from 'object-replace-mustache';
import sift from 'sift';

const defaultOptions = {
  skipHooks: false,
  params: void 0,
  name: "changesById",
  deleteParams: [],
  fetchBefore: false
};
const changesById = (cb, _options) => {
  const options = Object.assign({}, defaultOptions, _options);
  return async (context) => {
    if (shouldSkip("checkMulti", context)) {
      return context;
    }
    const pathBefore = getPath(options.name, true);
    if (context.type === "before") {
      const changes = await changesByIdBefore(context, options);
      if (!changes) {
        return context;
      }
      _set(context, pathBefore, changes);
    } else if (context.type === "after") {
      const itemsBefore = _get(context, pathBefore);
      const changes = await changesByIdAfter(context, itemsBefore, cb, options);
      if (!changes) {
        return context;
      }
      _set(context, getPath(options.name, false), changes);
    }
    return context;
  };
};
const updateMethods = ["update", "patch"];
const changesByIdBefore = async (context, _options) => {
  const options = Object.assign({}, defaultOptions, _options);
  let byId;
  if (context.method === "create" || !options.fetchBefore) {
    byId = {};
  } else if (updateMethods.includes(context.method) || context.method === "remove") {
    byId = await getOrFindById(context, options.params, { skipHooks: options.skipHooks });
  } else {
    return;
  }
  return byId;
};
const changesByIdAfter = async (context, itemsBefore, cb, _options) => {
  if (!itemsBefore) {
    return;
  }
  const options = Object.assign({}, defaultOptions, _options);
  const items = await resultById(context, options);
  if (!items) {
    return;
  }
  const itemsBeforeOrAfter = context.method === "remove" && options.fetchBefore ? itemsBefore : items;
  const changesById2 = Object.keys(itemsBeforeOrAfter).reduce(
    (result, id) => {
      if (options.fetchBefore && (context.method !== "create" && !itemsBefore[id] || context.method !== "remove" && !items[id])) {
        throw new Error("Mismatch!");
      }
      const before = itemsBefore[id];
      const item = items[id];
      result[id] = {
        before,
        item
      };
      return result;
    },
    {}
  );
  if (cb && typeof cb === "function") {
    await cb(changesById2, context);
  }
  return changesById2;
};
const getOrFindByIdParams = async (context, makeParams, options) => {
  if (context.id == null) {
    if (context.type === "before") {
      let params = copy(context.params);
      delete params.changesById;
      if (options?.deleteParams) {
        options.deleteParams.forEach((key) => {
          delete params[key];
        });
      }
      if (params.query?.$select) {
        delete params.query.$select;
      }
      params = Object.assign({ paginate: false }, params);
      params = typeof makeParams === "function" ? await makeParams(params, context) : params;
      return params;
    } else {
      if (!makeParams && !context.params.query?.$select) {
        return;
      }
      const itemOrItems = getItems(context);
      const idField = getIdField(context);
      if (!itemOrItems) {
        return;
      }
      const fetchedItems = Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems];
      const ids = fetchedItems.map((x) => x && x[idField]);
      let params = {
        query: {
          [idField]: { $in: ids }
        },
        paginate: false
      };
      params = makeParams ? await makeParams(params, context) : params;
      return params;
    }
  } else {
    if (context.type === "after" && !makeParams && !context.params.query?.$select) {
      return;
    }
    const query = Object.assign({}, context.params.query);
    delete query.$select;
    let params = Object.assign({}, context.params, { query });
    delete params.changesById;
    if (options?.deleteParams) {
      options.deleteParams.forEach((key) => {
        delete params[key];
      });
    }
    params = typeof makeParams === "function" ? await makeParams(params, context) : params;
    return params;
  }
};
const getOrFindById = async (context, makeParams, _options) => {
  const options = Object.assign({
    skipHooks: true,
    byId: true
  }, _options);
  let itemOrItems;
  const idField = getIdField(context);
  const params = await getOrFindByIdParams(context, makeParams, options);
  if (context.id == null) {
    const method = options.skipHooks ? "_find" : "find";
    itemOrItems = await context.service[method](params);
    itemOrItems = itemOrItems && (itemOrItems.data || itemOrItems);
  } else {
    const method = options.skipHooks ? "_get" : "get";
    itemOrItems = await context.service[method](context.id, params);
  }
  const items = !itemOrItems ? [] : Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems];
  if (options.byId) {
    return items.reduce((byId, item) => {
      const id = item[idField];
      byId[id] = item;
      return byId;
    }, {});
  } else {
    return items;
  }
};
const resultById = async (context, options) => {
  if (!context.result) {
    return {};
  }
  let items;
  let params = await getOrFindByIdParams(context, options.params, options);
  if (params) {
    const contextParams = Object.assign({}, context.params);
    delete contextParams.changesById;
    if (options?.deleteParams) {
      options.deleteParams.forEach((key) => {
        delete contextParams[key];
      });
    }
    if (_isEqual(params, context.params)) {
      params = null;
    }
  }
  if (context.method === "remove" || !params) {
    let itemOrItems = context.result;
    itemOrItems = Array.isArray(itemOrItems.data) ? itemOrItems.data : itemOrItems;
    items = Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems];
  } else {
    items = await getOrFindById(
      context,
      () => params,
      {
        skipHooks: options.skipHooks,
        byId: false
      }
    );
  }
  const idField = context.service.id;
  return items.reduce((byId, item) => {
    const id = item[idField];
    byId[id] = item;
    return byId;
  }, {});
};
const getIdField = (context) => {
  return context.service.options.id;
};
const getPath = (path, isBefore) => {
  if (isBefore) {
    if (typeof path === "string") {
      return `params.${path}.itemsBefore`;
    } else {
      return ["params", ...path, "itemsBefore"];
    }
  } else {
    if (typeof path === "string") {
      return `params.${path}`;
    } else {
      return ["params", ...path];
    }
  }
};

const trigger = (options) => {
  if (!options) {
    throw new Error("You should define subscriptions");
  }
  return async (context) => {
    checkContext(context, null, ["create", "update", "patch", "remove"], "trigger");
    if (context.type === "before") {
      return await triggerBefore(context, options);
    } else if (context.type === "after") {
      return await triggerAfter(context);
    }
  };
};
const triggerBefore = async (context, options) => {
  let subs = await getSubscriptions(context, options);
  if (!subs?.length) {
    return context;
  }
  if (!Array.isArray(context.data)) {
    const result = [];
    await Promise.all(
      subs.map(async (sub) => {
        if (!sub.action) {
          return;
        }
        sub.dataResolved = typeof sub.conditionsData === "function" ? await sub.conditionsData(context.data, context) : testCondition(context, context.data, sub.conditionsData);
        if (sub.dataResolved === false) {
          return;
        }
        sub.conditionsParamsResolved = typeof sub.conditionsParams === "function" ? await sub.conditionsParams(context.data, context) : testCondition(context, context.params, sub.conditionsParams);
        if (sub.conditionsParamsResolved === false) {
          return;
        }
        result.push(sub);
      })
    );
    subs = result;
  }
  if (!subs?.length) {
    return context;
  }
  for (const sub of subs) {
    if (checkConditions(sub)) {
      continue;
    }
    sub.paramsResolved = await getOrFindByIdParams(context, sub.params, { deleteParams: ["trigger"] });
    sub.identifier = JSON.stringify(sub.paramsResolved);
    if (context.params.changesById?.[sub.identifier]?.itemsBefore) {
      continue;
    }
    const before = await changesByIdBefore(context, {
      skipHooks: false,
      params: () => sub.params ? sub.paramsResolved : null,
      deleteParams: ["trigger"],
      fetchBefore: sub.fetchBefore || sub.conditionsBefore !== true
    });
    _set(context, ["params", "changesById", sub.identifier, "itemsBefore"], before);
  }
  setConfig(context, "subscriptions", subs);
  return context;
};
const triggerAfter = async (context) => {
  const subs = getConfig(context, "subscriptions");
  if (!subs?.length) {
    return context;
  }
  const now = /* @__PURE__ */ new Date();
  const promises = [];
  for (const sub of subs) {
    if (checkConditions(sub)) {
      continue;
    }
    const itemsBefore = context.params.changesById?.[sub.identifier]?.itemsBefore;
    let changesById;
    if (itemsBefore) {
      changesById = await changesByIdAfter(
        context,
        itemsBefore,
        null,
        {
          name: ["changesById", sub.identifier],
          params: sub.params,
          skipHooks: false,
          deleteParams: ["trigger"],
          fetchBefore: sub.fetchBefore
        }
      );
      _set(context, ["params", "changesById", sub.identifier], changesById);
    }
    changesById = context.params.changesById?.[sub.identifier];
    if (!changesById) {
      continue;
    }
    const changes = Object.values(changesById);
    for (const change of changes) {
      const { before } = change;
      const { item } = change;
      const changeForSub = change;
      const {
        conditionsResult,
        conditionsBefore
      } = sub;
      let mustacheView = {
        item,
        before,
        data: context.data,
        id: context.id,
        method: context.method,
        now,
        params: context.params,
        path: context.path,
        service: context.service,
        type: context.type,
        user: context.params?.user
      };
      if (sub.view) {
        if (typeof sub.view === "function") {
          mustacheView = await sub.view(mustacheView, { item: changeForSub, items: changes, subscription: sub, subscriptions: subs, context });
        } else {
          mustacheView = Object.assign(mustacheView, sub.view);
        }
      }
      sub.resultResolved = typeof conditionsResult === "function" ? await conditionsResult({ item, before }, context) : testCondition(mustacheView, item, conditionsResult);
      if (!sub.resultResolved) {
        continue;
      }
      sub.beforeResolved = typeof conditionsBefore === "function" ? await conditionsBefore({ item, before }, context) : testCondition(mustacheView, before, conditionsBefore);
      if (!sub.beforeResolved) {
        continue;
      }
      const _action = sub.action;
      const promise = _action(changeForSub, { subscription: sub, items: changes, context, view: mustacheView });
      if (sub.isBlocking) {
        promises.push(promise);
      }
    }
  }
  await Promise.all(promises);
  return context;
};
function setConfig(context, key, val) {
  context.params.trigger = context.params.trigger || {};
  context.params.trigger[key] = val;
}
function getConfig(context, key) {
  return context.params.trigger?.[key];
}
function checkConditions(sub) {
  return !sub.conditionsBefore && !sub.conditionsData && !sub.conditionsResult;
}
const defaultSubscription = {
  action: void 0,
  conditionsBefore: true,
  conditionsData: true,
  conditionsParams: true,
  conditionsResult: true,
  dataResolved: void 0,
  beforeResolved: void 0,
  resultResolved: void 0,
  isBlocking: true,
  method: void 0,
  params: void 0,
  service: void 0,
  view: void 0,
  paramsResolved: void 0,
  identifier: null,
  fetchBefore: false
};
const getSubscriptions = async (context, options) => {
  const _subscriptions = typeof options === "function" ? await options(context) : options;
  if (!_subscriptions) {
    return;
  }
  const subscriptions = Array.isArray(_subscriptions) ? _subscriptions : [_subscriptions];
  const subscriptionsResolved = subscriptions.map((x) => Object.assign({}, defaultSubscription, x));
  const { path, method } = context;
  return subscriptionsResolved.filter((sub) => {
    if (typeof sub.service === "string" && sub.service !== path || Array.isArray(sub.service) && !sub.service.includes(path)) {
      return false;
    }
    if (typeof sub.method === "string" && sub.method !== method || Array.isArray(sub.method) && !sub.method.includes(method)) {
      return false;
    }
    return true;
  });
};
const testCondition = (mustacheView, item, conditions) => {
  if (conditions === true) {
    return true;
  }
  conditions = copy(conditions);
  const transformedConditions = transformMustache(conditions, mustacheView);
  return sift(transformedConditions)(item) ? transformedConditions : false;
};

export { changesById, changesByIdAfter, changesByIdBefore, getOrFindByIdParams, trigger };
